{

=pod

=head1 NAME

Apache::XBEL

=head1 SUMMARY

 <Location />
  SetHandler  perl-script
  PerlHandler Apache::XBEL
  PerlSetVar  XbelPath     /path/to/yer-xbel-file.xbel
  PerlSetVar  XslPath      /path/to/apache-xbel.xsl
  PerlSetVar  CacheDir     /path/to/yer-cache-dir
 </Location> 

=head1 DESCRIPTION

Apache::XBEL is an Apache mod_perl handler that uses XSLT to transform XML Bookmarks Exchange Language (XBEL) files into exciting and foofy dynamic HTML documents. Documents are rendered as collapsible outlines and individual nodes may be viewed and bookmarked as unique pages, so you don't have to click through a gazillion nested leaves to find what you're looking for.

Once individual nodes/pages have been rendered, they are cached to reduce the load on the server. Cache files are updated whenever any of the widgets involved in the transformation are modified.

=cut

package Apache::XBEL;
use strict;

use vars qw ( $VERSION );
$VERSION = 1.1;

use Apache::Constants qw(:common :response);
use Apache::File;
use Apache::Log;
use Apache::Reload;
use Apache::URI;

use CGI;
use Digest::MD5;
use FileHandle;
use File::Basename;
use HTTP::BrowserDetect;

use XML::Generator;
use XML::LibXML;
use XML::LibXSLT;
use XML::XPath;

my $gObj_xpath;
my $gObj_xml_writer;

my $gObj_xml_parser;
my $gObj_xsl_parser;
my $gObj_xsl_stylesheet;
my $gObj_xsl_transformer;

my $gObj_browser;
my $gObj_cgi;
my $gObj_digest;

my $gVar_xbel_file;
my $gVar_xsl_file;

sub handler {
    my $apache = shift;

    my $root       = $apache->location;
    my $root_uri   = $apache->uri($root);
    $root_uri      =~ /($root)(.*)/;
    my $path       = $2;

    my $uri        = Apache::URI->parse($apache);
    my $uri_scheme = $uri->scheme;
    my $uri_host   = $uri->hostname;

    if (! ($root_uri =~ /(.*)\/$/)) {
	my $redirect = "$uri_scheme://$uri_host$root_uri/";
	$apache->headers_out->set("Location"=>$redirect);
	return REDIRECT;
    }

    &init($apache) || return SERVER_ERROR;

    $path =~ s/\/$//;
    $path =~ s/^\///;
    $root =  &basename($root);

    my ($cache_file,$exists) = &fetch_cache($apache,$path);

    if ($exists) {
	$apache->content_type("text/html");
	$apache->send_http_header();
	$apache->send_fd( Apache::File->new($cache_file) );
	return OK;
    }

    my @path        = split("/",$path);
    my @breadcrumbs = ("$root",split("/",$path));

    my $lookup      = &path2node(@path);
    my @nodes       = $gObj_xpath->findnodes($lookup);

    if (! @nodes) {
	$apache->log->error("Lookup for '$lookup' failed.\n");
	return NOT_FOUND;
    }

    my $new_cache = Apache::File->new();

    if (! $new_cache->open(">$cache_file")) { 
	$apache->log->error("Failed to open $cache_file for writing : $!\n"); 
	return SERVER_ERROR;
    }

    $new_cache = &lock_cache($apache,$new_cache) || return SERVER_ERROR;

    print $new_cache $gObj_xml_writer->xmldecl( version => "1.0",encoding=>"ISO-8859-1" );

    if ($nodes[0]->getName eq "xbel") {
	print $new_cache $nodes[0]->toString;
    }else{
	print $new_cache $gObj_xml_writer->xbel( &build_node($gObj_xml_writer,$nodes[0] ));
    }
    
    # Unlock
    close $new_cache;

    my $xmldoc  = $gObj_xml_parser->parse_file($cache_file);
    my $html    = $gObj_xsl_transformer->transform($xmldoc);

    $gObj_xsl_transformer->output_file($html,$cache_file);

    # Navbar
    my @nav_bar  = ();
    my $location = $apache->location();

    for (my $j = 0; $breadcrumbs[$j]; $j++) {
	next unless $breadcrumbs[$j+1];

	my $p = join("/",@breadcrumbs[1..$j]);
	my $h = "$uri_scheme://$uri_host$location/".join("/",@breadcrumbs[1..$j]);

	$p    = $gObj_xpath->findvalue(join("/",&path2node(split("/",$p)),"title"));
	$p    = $gObj_cgi->a({-href=>$h},$p);

	push(@nav_bar,$gObj_cgi->span({-class=>"navbar-item"},$p));
    }

    # Cleanup
    $gObj_xpath->cleanup();

    # Add navbar to file before returning content

    my $tmp_file    = "$cache_file.tmp";
    my $write_cache = Apache::File->new();

    $write_cache->open(">$tmp_file");

    # Lock the file
    $write_cache = &lock_cache($apache,$write_cache) || return SERVER_ERROR;

    # For some reason, when I use the Apache::File
    # methods to do this, Apache never stops printing
    # white space. Dunno...
    my $read_cache = FileHandle->new();

    if (! $read_cache->open("<$cache_file")) {
	$apache->log->error("Failed to open $cache_file for reading : $!\n");
	return SERVER_ERROR;
    }
    
    while (<$read_cache>) {
	if ($_ =~ /(.*)(<body.*?>)(<.*)/) { 
	    $_ = join("",$1,$2,$gObj_cgi->div({-class=>"navbar"},join("\n",@nav_bar)),$3); 
	}
	print $write_cache $_;
    }

    # Make sure we close the file 
    # and give up the lock
    $write_cache->close;
    $read_cache->close();

    if (! rename $tmp_file, $cache_file) {
	$apache->log->error("There was an error renaming the $tmp_file : $!");
	return SERVER_ERROR;
    }

    # Send the stupid file, already
    $apache->content_type("text/html");
    $apache->send_http_header();
    $apache->send_fd(new Apache::File($cache_file));

    return OK;
}

sub cleanup_handler {
    my $apache = shift;
    return 1;
}

sub init {
    my $apache = shift;

    $gVar_xbel_file = $apache->dir_config("XbelPath") ||
	
	&{
	    $apache->log->error("Unable to locate '$gVar_xbel_file'.\n");
	    return 0;
	};

#    if (&has_valid_DOM) {
#	$gVar_xsl_file = $apache->dir_config("XslPath");
#    } 

#    else { 
#	$gVar_xsl_file = $apache->dir_config("XslPathNoDOM");
#    }

    $gVar_xsl_file = $apache->dir_config("XslPath");
    
    if (! -e $gVar_xsl_file) {
	$apache->log->error("Unable to locate '$gVar_xsl_file'.\n");
	return 0;
    }

    if (! -e $apache->dir_config("CacheDir")) {
	$apache->log->error("Unable to locate the cache directory.\n");
	return 0;
    }
    
    $gObj_xpath = new XML::XPath(filename=>$gVar_xbel_file) ||

	&{
	    $apache->log->alert("Unable to create XPath object : $!\n");
	    return 0;
	};

#    $gObj_browser ||= new HTTP::BrowserDetect ||

#	&{
#	    $apache->log->alert("Unable to create BrowserDetect object : $!\n");
#	    return 0;
#	};

    $gObj_xml_writer ||= new XML::Generator(pretty=> 2,conformance=>"strict") ||

	&{
	    $apache->log->alert("Unable to create XML::Generator object. $!");
	    return 0;
	};

    $gObj_xml_parser ||= XML::LibXML->new();

    if (! $gObj_xml_parser) {
	$apache->log->alert("Unable to create XML::LibXML object. $!");
	return 0;
    } 

    else {
	$gObj_xml_parser->validation(1);
    }

    $gObj_xsl_parser ||= XML::LibXSLT->new() ||

	&{
	    $apache->log->alert("Unable to create XML::LibXSLT object. $!");
	    return 0;
	};

    $gObj_xsl_stylesheet ||= $gObj_xml_parser->parse_file($gVar_xsl_file) ||

	&{
	    $apache->log->alert("Failed to parse file $gVar_xsl_file");
	    return 0;
	};

    $gObj_xsl_transformer ||= $gObj_xsl_parser->parse_stylesheet($gObj_xsl_stylesheet) ||
	
	&{
	    $apache->log->alert("Failed to parse stylesheet object.");
	    return 0;
	};

    $gObj_cgi ||= new CGI ||
	
	&{
	    $apache->log->alert("Unable to create CGI object : $!\n");
	    return 0;
	};

    $gObj_digest = new Digest::MD5->new() ||

	&{
	    $apache->log->alert("Unable to create MD5 object : $!\n");
	    return 0;
	};

    return 1;
}

sub fetch_cache {
    my $apache    = shift;
    my $path      = shift;

    # First determine cache name
    $gObj_digest->add(join("/",$gVar_xbel_file,$path));
    my $cache = join("/",$apache->dir_config("CacheDir"),$gObj_digest->hexdigest());

    $apache->log->debug("Cache is $cache");

    # Check for existence
    if (! -e $cache) {
	$apache->log->debug("Cachefile '$cache' does not exist.");
	return ($cache,0);
    }

    # Is older ?
    my $cache_mtime = (stat($cache))[9];

    if ((stat(__FILE__))[9] > $cache_mtime ) {
	$apache->log->debug("Cache is out of sync with handler.");
	return ($cache,0);
    }

    if ((stat($gVar_xbel_file))[9] > $cache_mtime ) {
	$apache->log->debug("Cache is out of sync with XBEL file.");
	return ($cache,0);
    }

    if ((stat($gVar_xsl_file))[9] > $cache_mtime) {
	$apache->log->debug("Cache is out of sync with stylesheet.");

	$gObj_xsl_stylesheet  = $gObj_xml_parser->parse_file($gVar_xsl_file);
	$gObj_xsl_transformer = $gObj_xsl_parser->parse_stylesheet($gObj_xsl_stylesheet);
	
	return ($cache,0);
    }

    return ($cache,1);
}

sub has_valid_DOM {
    my $apache = shift;
    # Check browser here
    return 1;
}

sub path2node {
    my @path = @_;
    map { $_ = "folder[\@id=\"$_\"]"; } @path;
    return join("/","","xbel",@path);
}

sub build_node {
    my $xbel = shift;
    my $node = shift;
    
    my $author   = "";
    my $desc     = "";

    my @children = $node->getChildNodes();
    my @body;

    for (my $i = 1; $children[$i]; $i++) {
	my $child = $node->getChildNode($i)->getName();
	push(@body,\$node->getChildNode($i)->toString());
    }
    
    my $n = join("",(
		     $xbel->info($xbel->metadata({ owner => $author })),
		     $xbel->desc($desc),
		     map { $$_ } @body,
		     ));
    
    return $n;
}

sub lock_cache {
    my $apache = shift;
    my $fh     = shift;

    my $success = 0;
    my $tries   = 0;

    while ($tries++ < 10) {
	return $fh if ($success = flock($fh,2));
	sleep(1);
    }

    $apache->log->error("Failed to lock file for writing.");
    return undef;
}

=pod

=head1 TO DO

=over 4

=item

Add support for Apache::Filter.

=item

Write a plain vanila stylesheet and write DOM related functions.

=back

=head1 VERSION

1.1

=head1 DATE

October 18, 2001

=head1 CHANGES

=head2 1.1

=over 4

=item 

Cleaned up naming conventions

=item

Replace XML::Sablotron with XML::LibXSLT

=item

Updated POD

=head2 1.0

=over 4

=item

Initial revision

=back

=head1 AUTHOR

Aaron Straup Cope

=head1 SEE ALSO

http://pyxml.sourceforge.net/topics/xbel/

=head1 LICENSE

Copyright 2001 Aaron Straup Cope.

This is free software, you may use it and distribute it under the same terms as Perl itself.

=cut

return 1;

}
